import os
import warnings
from typing import Optional

from pypaq.lipytools.logger import set_logger
from pypaq.mpython.devices import DevicesParam, get_devices
from pypaq.neuralmess.get_tf import tf

# init function for every TF.py script:
# - sets low verbosity of TF
# - starts logger
# - manages TF devices
def nestarter(
        log_folder: Optional[str]=  '_log', # for None doesn't log
        custom_name: str or None=   None,   # set custom logger name, for None uses default
        devices: DevicesParam=      -1,     # False to not manage TF devices
        verb=                       1,
        silent_error=               False): # turns off any TF errors, be careful

    tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)
    warnings.filterwarnings('ignore')
    os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3' if silent_error else '2'

    if log_folder: set_logger(log_folder=log_folder, custom_name=custom_name, verb=verb) # set logger
    if devices is not False: return get_devices(devices=devices, namespace='TF1')


"""

 2020 (c) piteren

    baseline DQN graph for NEModel (TF)

"""

from pypaq.neuralmess.base_elements import tf
from pypaq.neuralmess.layers import lay_dense


def dqn_graph(
        name=               'qnn',
        num_actions: int=   4,
        observation_width=  4,
        hidden_layers=      (12,),
        seed=               121):

    with tf.variable_scope(name):

        observations_PH = tf.placeholder(   # observations input vector
            shape=  [None,observation_width],
            dtype=  tf.float32,
            name=   'observations')
        enum_actions_PH = tf.placeholder(   # enumerated action indexes (0,1),(1,3),(2,0),..
            shape=  [None,2],
            dtype=  tf.int32,
            name=   'enum_actions')
        gold_QV_PH = tf.placeholder(        # gold_QV
            shape=  [None],
            dtype=  tf.float32,
            name=   'gold_QV')

        layer = tf.keras.layers.LayerNormalization(axis=-1)(observations_PH)
        for i in range(len(hidden_layers)):
            layer = lay_dense(
                input=      layer,
                name=       f'hidden_layer_{i}',
                units=      hidden_layers[i],
                activation= tf.nn.relu,
                seed=       seed)
            layer = tf.keras.layers.LayerNormalization(axis=-1)(layer)

        # QV for all actions (for given input(state))
        output = lay_dense(
            input=      layer,
            units=      num_actions,
            activation= None,
            seed=       seed)

        pred_qv = tf.gather_nd(output, indices=enum_actions_PH)

        # loss on predicted vs next, we want predicted to match next
        loss = tf.losses.mean_squared_error(
            labels=         gold_QV_PH,
            predictions=    pred_qv,
            reduction=      tf.compat.v1.losses.Reduction.SUM_OVER_BATCH_SIZE)

    return {
        'gold_QV_PH':       gold_QV_PH,
        'observations_PH':  observations_PH,
        'enum_actions_PH':  enum_actions_PH,
        'output':           output,
        'loss':             loss}


"""

 2018 (c) piteren

"""

import string


# separates punctuation with spaces
def pretokenize_punct(text):
    newText = ''
    for ix in range(len(text)):
        c = text[ix]
        if c in string.punctuation:
            newText += f' {c} '
        else: newText += c
    return newText

# whitespace tokenizer
def whitspace_tokenizer(text):
    return text.split()

# whitespace tokenizer
def whitspace_normalize(text):
    text_list = whitspace_tokenizer(text)
    return ' '.join(text_list)


if __name__ == '__main__':

    text = 'Just   how is  Hillary      Kerr, the founder of a digital media company in Los Angeles? She can tell you what song was playing five years ago on the jukebox at the bar where she somewhat randomly met the man who became her husband.'

    print(f'>{whitspace_normalize(text)}<')

    """
    print(pretokenize_punct(text))
    print('\n***words:')
    for word in tokenize_words(text, tokenizer=None): print(' >%s<' % word)
    print('\n***sentences:')
    for sent in tokenize_sentences(text): print(' >%s<' % sent)
    """


"""

 2021 (c) piteren

    OMPRunner - Object based Multi Processing Runner

        Tasks may be given for OMPRunner with process() method - one dict after another or in packages (List[dict]).
        Results may be received with two get methods (single or all) and by default will be ordered with tasks order.
        Result returned by RunningWorker may be Any.
        OMPRunner needs to be manually closed with exit().

        OMPRunner processes given tasks with InternalProcessor (IP) that guarantees non-blocking interface of OMPRunner.
        Tasks are processed directly by RunningWorker class objects that are managed by IP.

        RunningWorker must be inherited and its process() implemented.
        process() takes task via **kwargs and returns task result.
        There are two main policies of RunningWorker lifecycle:
            1st - RunningWorker is closed after processing some task (1 typically but may be N)
            2nd - RunningWorker is closed only when crashes or with the OMP exit
            Each policy has job specific pros and cons. By default, second is activated with 'rw_lifetime=None'.
            Pros of 'rw_lifetime=None' are:
            + all RunningWorkers are initialized once while OMP inits. It saves a time.
            Cons are:
            - memory kept by the RunningWorker may grow with the time (while processing many tasks).

"""

from abc import ABC, abstractmethod
from collections import deque
from inspect import getfullargspec
import os
import psutil
import time
from typing import Any, List, Dict, Optional, Tuple, Union

from pypaq.lipytools.pylogger import get_pylogger, get_hi_child
from pypaq.lipytools.moving_average import MovAvg
from pypaq.mpython.devices import DevicesParam, get_devices
from pypaq.mpython.mptools import QMessage, ExSubprocess, Que


class OMPRException(Exception):
    pass

# result returned in case of RWW exception
class ResultOMPRException:
    def __init__(self, exception_msg:str, task:dict):
        self.exception_msg = exception_msg
        self.task = task

# interface of running object - processes task given with kwargs and returns result
class RunningWorker(ABC):

    @ abstractmethod # processing method to be implemented
    def process(self, **kwargs) -> Any: pass


class OMPRunner:

    # Internal Processor of OMPRunner
    class InternalProcessor(ExSubprocess):

        POISON_MSG = QMessage(type='poison', data=None)

        # RWW wraps RunningWorker with Exception Managed Subprocess
        class RWWrap(ExSubprocess):

            def __init__(
                    self,
                    ique: Que,
                    oque: Que,
                    id: int,
                    rw_class: type(RunningWorker),
                    rw_init_kwargs: Dict,
                    raise_unk_exception,
                    logger):
                ExSubprocess.__init__(
                    self,
                    ique=                   ique,
                    oque=                   oque,
                    id=                     id,
                    raise_unk_exception=    raise_unk_exception,
                    logger=                 logger)
                self.rw_class = rw_class
                self.rw_init_kwargs = rw_init_kwargs
                self.logger.info(f'> RWWrap({self.id}) initialized')

            # loop for processing RW tasks, this method will be run within subprocess (ExSubprocess)
            def subprocess_method(self):

                self.logger.info(f'> RWWrap ({self.id}) pid: {os.getpid()} inits RunningWorker')
                rwo = self.rw_class(**self.rw_init_kwargs)
                self.logger.info(f'> RWWrap ({self.id}) starts process loop..')

                while True:
                    ompr_msg: QMessage = self.ique.get()
                    if ompr_msg.type == 'break': break
                    if ompr_msg.type == 'hold_check':self.oque.put(QMessage(type='hold_ready', data=None))
                    if ompr_msg.type == 'task':
                        result = rwo.process(**ompr_msg.data['task'])
                        self.oque.put(QMessage(
                            type=   'result',
                            data=   {
                                'rww_id':   self.id,
                                'task_ix':  ompr_msg.data['task_ix'],
                                'task_try': ompr_msg.data['task_try'],
                                'result':   result}))

                self.logger.info(f'> RWWrap (id: {self.id}) finished process loop')


        def __init__(
                self,
                ique: Que,                              # tasks and other messages
                oque: Que,                              # results and exception_results
                rw_class: type(RunningWorker),
                rw_init_kwargs: Optional[Dict],
                rw_lifetime: Optional[int],
                devices: DevicesParam,
                ordered_results: bool,
                task_timeout: Optional[float],
                restart_ex_tasks: bool,
                log_exceptions: bool,
                raise_RWW_exception: bool,
                report_delay: Optional[int],
                logger):

            self.rw_class = rw_class
            self.ip_name = f'InternalProcessor_for_{self.rw_class.__name__}' # INFO: .name conflicts with Process.name

            self.logger = logger
            self.logger.info(f'*** {self.ip_name} *** (pid: {os.getpid()}) inits..')

            # adds to InternalProcessor Exception Managed Subprocess properties
            ExSubprocess.__init__(
                self,
                ique=           ique,
                oque=           oque,
                id=             self.ip_name,
                logger=         self.logger)

            self.que_RW = Que() # here OMP receives messages from RW ('result' or 'ex_..'/exception)

            if not rw_init_kwargs: rw_init_kwargs = {}
            self.rw_lifetime = rw_lifetime

            devices = get_devices(devices=devices, namespace=None)

            dev_param_name = None
            pms = getfullargspec(self.rw_class).args
            if 'devices' in pms: dev_param_name = 'devices'
            if 'device' in pms: dev_param_name = 'device'

            self.ordered_results = ordered_results
            self.task_timeout = task_timeout
            self.restart_ex_tasks = restart_ex_tasks
            self.log_exceptions = log_exceptions
            self.raise_RWW_exception = raise_RWW_exception
            self.report_delay = report_delay

            # prepare RWW dictionary, keeps RWW id, RW init kwargs, RWW object
            self.rwwD: Dict[int, Dict] = {}  # {rww.id: {'rw_init_kwargs':{}, 'rww':RWWrap}}
            for id, dev in enumerate(devices):
                kwD = {}
                kwD.update(rw_init_kwargs)
                if dev_param_name: kwD[dev_param_name] = dev
                self.rwwD[id] = {
                    'rw_init_kwargs':   kwD,
                    'rww':              None}

        # builds and starts single RWWrap
        def _build_and_start_RWW(self, id:int):
            assert self.rwwD[id]['rww'] is None
            self.rwwD[id]['rww'] = OMPRunner.InternalProcessor.RWWrap(
                ique=                   Que(),
                oque=                   self.que_RW,
                id=                     id,
                rw_class=               self.rw_class,
                rw_init_kwargs=         self.rwwD[id]['rw_init_kwargs'],
                raise_unk_exception=    self.raise_RWW_exception,
                logger=                 get_hi_child(self.logger))
            self.rwwD[id]['rww'].start()
            self.logger.debug(f'> {self.ip_name} built and started RWWrap id: {id}..')

        def build_and_start_allRWW(self):
            n_started = 0
            for id in self.rwwD:
                if self.rwwD[id]['rww'] is None:
                    self._build_and_start_RWW(id)
                    n_started += 1
            self.logger.info(f'> {self.ip_name} built and started {n_started} RunningWorkers')

        # kills single RWWrap
        def _kill_RWW(self, id:int):
            self.rwwD[id]['rww'].kill()
            while True: # we have to flush the RW ique
                ind = self.rwwD[id]['rww'].ique.get_if()
                #if ind: print(f'@@@ got ind of RW {id}: {ind}')
                if not ind: break
            self.rwwD[id]['rww'].join()
            self.rwwD[id]['rww'] = None
            self.logger.debug(f'> {self.ip_name} killed and joined RWWrap id: {id}..')

        def _kill_allRWW(self):
            for id in self.rwwD:
                if self.rwwD[id]['rww'] is not None:
                    self._kill_RWW(id)
            self.logger.info(f'> {self.ip_name} killed and joined {len(self.rwwD)} RunningWorkers')

        # this method checks if all RunningWorkers are ready to process tasks
        def hold_till_allRWW_ready(self):
            self.logger.info(f'> hold: {self.ip_name} is checking RW readiness..')
            for id in self.rwwD:
                if self.rwwD[id]['rww'] is None:
                    print(f'WARNING: some RW are not started, cannot hold!!!')
                    return
            for id in self.rwwD: self.rwwD[id]['rww'].ique.put(QMessage(type='hold_check', data=None))
            for _ in self.rwwD: self.que_RW.get()
            self.logger.info(' > hold: all RW are ready')

        # returns string with information about subprocesses
        def _get_RWW_info(self) -> str:

            ip_id = os.getpid()
            ip_mem = int(psutil.Process(ip_id).memory_info().rss / 1024 ** 2)
            vm = psutil.virtual_memory()
            used = vm.used / 1024 ** 3

            num_all = len(self.rwwD)
            num_alive = sum([1 for rww_id in self.rwwD if self.rwwD[rww_id]['rww'] is not None and self.rwwD[rww_id]['rww'].alive])
            num_closed = sum([1 for rww_id in self.rwwD if self.rwwD[rww_id]['rww'] is not None and self.rwwD[rww_id]['rww'].closed])
            alive_info = f'{num_all}= alive:{num_alive} closed:{num_closed}'

            rww_mem = [int(psutil.Process(self.rwwD[id]['rww'].pid).memory_info().rss / 1024 ** 2) for id in self.rwwD if self.rwwD[id]['rww'].alive]
            rww_mem.sort(reverse=True)

            tot_mem = ip_mem + sum(rww_mem)
            s = f'# {self.ip_name} mem: {ip_mem}MB, omp+sp/used: {tot_mem/1024:.1f}/{used:.1f}GB ({int(vm.percent)}%VM)\n'
            if len(rww_mem) > 6: s += f'# subproc: {rww_mem[:3]}-{int(sum(rww_mem)/len(rww_mem))}-{rww_mem[-3:]} ({alive_info})'
            else:                s += f'# subproc: {rww_mem} ({alive_info})'
            return s

        # main loop of OMP_IP, run by ExSubprocess
        def subprocess_method(self):

            self.logger.info(f'> {self.ip_name} loop starts with {len(self.rwwD)} subprocesses (RWW)')
            self.build_and_start_allRWW()

            next_task_ix = 0                                # next task index (index of task that will be processed next)
            task_result_ix = 0                              # index of task result that should be put to self.results_que now
            rww_ntasks = {k: 0 for k in self.rwwD.keys()}   # number of tasks processed by each RWW since restart

            iv_time = time.time()                           # interval report time
            iv_n_tasks = 0                                  # number of tasks finished since last interval
            n_tasks_processed = 0                           # number of tasks processed (total)
            speed_cache = MovAvg(factor=0.2)                # speed moving average
            tasks_que = deque()                             # que of (task_ix, task_try, task) to be processed (received from the self.tasks_que or restarted after exception)
            resources = list(self.rwwD.keys())              # list [rww_id] of all available (not busy) resources
            rww_tasks: Dict[int, Tuple[int,int,Any]] = {}   # {rww.id: (task_ix,task_try,task)}
            task_times: Dict[int, float] = {}               # {rww.id: time} here we note time when RWW got current task, it is a dict but we also use property that order is maintained
            killed_tasks: Dict[int, Tuple[int,int]] = {}    # {rww_id: (task_ix,task_try)} here we note tasks that may come from RWW that were killed, we need to get rid of them

            resultsD: Dict[int, Any] = {}                   # results dict {task_ix: result(data)}
            while True:

                break_loop = False # flag to break this loop

                # to allow poison to be received while tasks still present
                msg = self.ique.get_if()
                if msg:
                    if msg.type == 'poison':
                        break_loop = True
                    if msg.type == 'tasks':
                        for task in msg.data:
                            tasks_que.append((next_task_ix, 0, task))
                            next_task_ix += 1

                if not tasks_que and len(resources) == len(self.rwwD):
                    msg = self.ique.get() # wait here for task or poison
                    if msg.type == 'poison':
                        break_loop = True
                    if msg.type == 'tasks':
                        for task in msg.data:
                            tasks_que.append((next_task_ix, 0, task))
                            next_task_ix += 1

                # try to flush the que
                if not self.ique.empty() or self.ique.qsize():
                    self.logger.debug(f'ique is empty - {self.ique.empty()}, it has {self.ique.qsize()} objects, trying to flush in the 2nd loop')
                    while True:
                        msg = self.ique.get_if()
                        if msg:
                            if msg.type == 'poison':
                                break_loop = True
                                break
                            if msg.type == 'tasks':
                                for task in msg.data:
                                    tasks_que.append((next_task_ix, 0, task))
                                    next_task_ix += 1
                        else: break

                if break_loop:
                    # all RWW have to be killed here, from the loop
                    # we want to kill them because it is quicker than waiting for them till finish tasks
                    # - we do not need their results anymore
                    self._kill_allRWW()
                    break

                self.logger.debug(f'>> free resources: {len(resources)}, tasks_que len: {len(tasks_que)}, ique.qsize: {self.ique.qsize()}')

                # put all needed resources into work
                while resources and tasks_que:

                    rww_id = resources.pop(0) # take first free resource

                    # eventually restart RWW (lifetime limit)
                    if self.rw_lifetime and rww_ntasks[rww_id] >= self.rw_lifetime:
                        self.logger.debug(f'>> restarting RWWrap id: {rww_id}...')
                        self._kill_RWW(rww_id)
                        self._build_and_start_RWW(rww_id)
                        rww_ntasks[rww_id] = 0

                    # get first task, prepare and put message for RWWrap
                    task_ix, task_try, task = tasks_que.popleft()
                    msg = QMessage(
                        type=   'task',
                        data=   {'task_ix':task_ix, 'task_try':task_try, 'task':task})
                    self.rwwD[rww_id]['rww'].ique.put(msg)
                    rww_tasks[rww_id] = (task_ix, task_try, task)
                    task_times[rww_id] = time.time()

                    self.logger.debug(f'>> put task {task_ix} for RWWrap({rww_id})')

                ### flush que_RW (get messages from RWWraps)
                rww_msgL: List[QMessage] = []

                ### get at least one
                # do it in a timeout loop to check for over-timed RWW
                if self.task_timeout:
                    msg = None
                    while not msg:
                        msg = self.que_RW.get_timeout(timeout=self.task_timeout/5)

                        oldest_rww_id = list(task_times.keys())[0]
                        if time.time() - task_times[oldest_rww_id] > self.task_timeout:
                            self.rwwD[oldest_rww_id]['rww'].kill()      # kill him
                            task_ix, task_try, _ = rww_tasks[oldest_rww_id] # get his task data
                            killed_tasks[oldest_rww_id] = (task_ix, task_try)
                            if msg: rww_msgL.append(msg)                # do not lose good msg
                            # prepare msg for killed RWW (killed RRW does not return message after kill)
                            msg = QMessage(
                                type=   f'ex_timeout_killed, ExSubprocess id: {oldest_rww_id}',
                                data=   oldest_rww_id)  # return ID here to allow process identification

                else: msg = self.que_RW.get()

                # get more (ready)
                while msg:
                    rww_msgL.append(msg)
                    msg = self.que_RW.get_if()
                self.logger.debug(f'>> received {len(rww_msgL)} messages from RWWraps')

                # eventually get rid of task from killed RWW
                # INFO: it is possible that killed process returned its result in the meanwhile (while processing kill)
                # eventually clean_up killed_tasks
                cleaned_rww_msgL = []   # new, clean lost
                clean_rww_id = []       # clean those RWW from killed_tasks
                for msg in rww_msgL:
                    msg_ok = True
                    for rww_id in killed_tasks:
                        if msg.type == 'result' and msg.data['rww_id'] == rww_id:
                            clean_rww_id.append(rww_id)
                            if msg.data['task_ix'] == killed_tasks[rww_id][0] and msg.data['task_try'] == killed_tasks[rww_id][1]:
                                msg_ok = False
                    if msg_ok: cleaned_rww_msgL.append(msg)
                for rww_id in clean_rww_id:
                    killed_tasks.pop(rww_id)
                rww_msgL = cleaned_rww_msgL

                # process all RWW messages
                for msg in rww_msgL:

                    if msg.type == 'result':

                        rww_id = msg.data['rww_id']
                        rww_ntasks[rww_id] += 1

                        task_ix, task_try, _ = rww_tasks.pop(rww_id) # ..could also be get from msg.data
                        task_times.pop(rww_id)

                        iv_n_tasks += 1
                        n_tasks_processed += 1
                        self.logger.debug(f'>> got result of task_ix: {task_ix} try: {task_try}')

                        res_msg = QMessage(type='result', data=msg.data['result'])
                        if self.ordered_results: resultsD[task_ix] = res_msg
                        else: self.oque.put(res_msg)

                        resources.append(rww_id)

                    # RWWrap exception
                    else:
                        if 'ex_' not in msg.type: raise OMPRException('ERR: unknown RWWrap message received!')

                        rww_id = msg.data
                        task_ix, task_try, task = rww_tasks.pop(rww_id)
                        task_times.pop(rww_id)

                        if self.log_exceptions:
                            self.logger.warning(f'> {self.ip_name} received exception message: {msg.type}, not finished task_ix: {task_ix}, recreating RWWrap..')

                        # close rww
                        rww = self.rwwD[rww_id]['rww']
                        if rww is not None: # if RWW was killed it is None then
                            rww.join() # we cannot kill that process since only alive process can be killed
                            rww.close()
                            self.rwwD[rww_id]['rww'] = None

                        # rebuild RWW and put to resources
                        self._build_and_start_RWW(rww_id)
                        rww_ntasks[rww_id] = 0
                        resources.append(rww_id)

                        if self.restart_ex_tasks:
                            if self.log_exceptions:
                                self.logger.warning(f'>> putting task again to the tasks_que..')
                            tasks_que.appendleft((task_ix, task_try+1, task))

                        else:
                            if self.log_exceptions:
                                self.logger.warning(f'>> returning exception result..')

                            res_msg = QMessage(
                                type=   'exception_result',
                                data=   ResultOMPRException(exception_msg=msg.type, task=task))
                            if self.ordered_results: resultsD[task_ix] = res_msg
                            else: self.oque.put(res_msg)

                    # flush resultsD
                    while task_result_ix in resultsD:
                        self.oque.put(resultsD.pop(task_result_ix))
                        task_result_ix += 1

                if self.report_delay is not None and time.time()-iv_time > self.report_delay:
                    iv_speed = iv_n_tasks/((time.time()-iv_time)/60)
                    speed = speed_cache.upd(iv_speed)
                    if speed != 0:
                        if speed > 10:      speed_str = f'{int(speed)} tasks/min'
                        else:
                            if speed > 1:   speed_str = f'{speed:.1f} tasks/min'
                            else:           speed_str = f'{1 / speed:.1f} min/task'
                        n_tasks_left = len(tasks_que)
                        est = n_tasks_left / speed
                        n_total_tasks = n_tasks_processed + n_tasks_left
                        progress = n_tasks_processed / n_total_tasks
                        self.logger.info(f'> ({progress * 100:4.1f}% left:{n_tasks_left}/{n_total_tasks} {time.strftime("%H:%M:%S")}) speed: {speed_str}, EST:{est:.1f}min')
                    else: self.logger.info(f'> processing speed unknown yet..')
                    iv_time = time.time()
                    iv_n_tasks = 0
                    self.logger.debug(self._get_RWW_info())

        def after_exception_handle_run(self):
            self._kill_allRWW()
            self.logger.debug(f'> {self.ip_name} killed all RW after exception occurred')

        def get_num_RWW(self):
            return len(self.rwwD)

        # method to call out of the process (to exit it)
        def exit(self) -> None:

            if self.alive:
                self.ique.put(self.POISON_MSG)

                while self.alive:
                    # flush the oque
                    while True:
                        res = self.oque.get_if()
                        if res is None: break
                    self.join(timeout=0.0001)

    def __init__(
            self,
            rw_class: type(RunningWorker),          # RunningWorker class that will run() given tasks
            rw_init_kwargs: Optional[Dict]= None,   # RunningWorker __init__ kwargs
            rw_lifetime: Optional[int]=     None,   # RunningWorker lifetime, for None or 0 is unlimited, for N <1,n> each RW will be restarted after processing N tasks
            devices: DevicesParam=          'all',
            name: str=                      'OMPRunner',
            ordered_results: bool=          True,   # returns results in the order of tasks
            task_timeout: Optional[float]=  None,   # (sec) timeout for task, if task processing time exceeds time, process will be killed and task not restarted, INFO: will then return result of type OMPRException
            restart_ex_tasks: bool=         True,   # restarts tasks that caused exception, INFO: for False may return result of type OMPRException
            log_exceptions: bool=           True,   # allows to log exceptions
            raise_RWW_exception: bool=      False,  # forces RWW to raise exceptions (.. all but KeyboardInterrupt)
            report_delay: Union[int,str]=   'auto', # num sec between speed_report, 'auto' uses loglevel, for 'none' there is no speed report
            logger=                         None,
            loglevel=                       20):

        self.omp_name = name

        if not logger:
            logger = get_pylogger(
                name=       self.omp_name,
                add_stamp=  False,
                folder=     None,
                level=      loglevel)
        self.logger = logger

        self.logger.info('*** OMPRunner *** inits..')
        self.logger.info(f'> name:     {self.omp_name}')
        self.logger.info(f'> pid:      {os.getpid()}')
        self.logger.info(f'> rw_class: {rw_class.__name__}')

        self._tasks_que = Que()             # que of tasks to be processed
        self._results_que = Que()           # que of ready results
        self._n_tasks_received: int = 0     # number of tasks received from user till now
        self._n_results_returned: int = 0   # number of results returned to user till now

        if report_delay == 'none': report_delay = None
        if report_delay == 'auto': report_delay = 30 if loglevel>10 else 10

        self._internal_processor = OMPRunner.InternalProcessor(
            ique=                   self._tasks_que,
            oque=                   self._results_que,
            rw_class=               rw_class,
            rw_init_kwargs=         rw_init_kwargs if rw_init_kwargs else {},
            rw_lifetime=            rw_lifetime,
            devices=                devices,
            ordered_results=        ordered_results,
            task_timeout=           task_timeout,
            restart_ex_tasks=       restart_ex_tasks,
            log_exceptions=         log_exceptions,
            raise_RWW_exception=    raise_RWW_exception,
            report_delay=           report_delay,
            logger=                 self.logger)
        self._internal_processor.start()

    # (not blocking) takes tasks for processing, starts processing, does not return anything
    def process(self, tasks: dict or List[dict]):
        if type(tasks) is dict: tasks = [tasks]
        self._tasks_que.put(QMessage(type='tasks', data=tasks))
        self._n_tasks_received += len(tasks)

    # returns single result, may block or not
    def get_result(self, block=True) -> Optional[Any]:
        if self._n_results_returned == self._n_tasks_received:
            self.logger.info(f'OMPRunner get_result() returns None since already returned all results (for all given tasks: n_results_returned == n_tasks_received)')
            return None
        else:
            if block:   msg = self._results_que.get()
            else:       msg = self._results_que.get_if()
            if msg:
                self._n_results_returned += 1
                return msg.data
            return None

    # returns results of all tasks put up to NOW
    def get_all_results(self, pop_ex_results=False) -> List[Any]:
        results = []
        n_results = self._n_tasks_received - self._n_results_returned
        while len(results) < n_results:
            results.append(self.get_result(block=True))
        if pop_ex_results:
            results = [r for r in results if type(r) is not ResultOMPRException]
        return results

    def get_tasks_stats(self) -> Dict[str,int]:
        return {
            'n_tasks_received':     self._n_tasks_received,
            'n_results_returned':   self._n_results_returned}

    def get_num_workers(self):
        return self._internal_processor.get_num_RWW()

    def exit(self):
        if self._n_results_returned != self._n_tasks_received:
            self.logger.warning(f'{self.omp_name} exits while not all results were returned to user!')
        self._internal_processor.exit()
        self.logger.info(f'> {self.omp_name}: internal processor stopped, {self.omp_name} exits.')



"""

 2019 (c) piteren

"""

import numpy as np
from typing import List

from pypaq.neuralmess.get_tf import tf
from pypaq.neuralmess.base_elements import my_initializer, flatten_LOTens


# residual (advanced) connection for (any) layer
def lay_res(
        lay_in,                     # layer input
        lay_out,                    # layer output
        name=           'residual',
        use_RCW=        False,      # use residual connection weights
        use_PWRCW=      False,      # pointwise weights
        match_dims=     True):      # concatenates zeros to input when thinner

    # TODO: not working for higher dimm tensors
    with tf.variable_scope(name):

        output = lay_out
        iW = int(lay_in.shape[-1])
        oW = int(output.shape[-1])
        matchedDims = iW == oW

        # pad input with zeros to match dimension of output
        if iW < oW and match_dims:
            lay_in = tf.pad(
                tensor=     lay_in,
                paddings=   tf.constant([[0,0],[0,oW-iW]]))
            matchedDims = True

        if matchedDims:
            if use_RCW:
                if use_PWRCW:   shape = [oW]
                else:           shape = []

                convRCW = tf.get_variable(
                    name=           'rcw',
                    shape=          shape,
                    initializer=    tf.constant_initializer(0))

                output = lay_in * (1 - tf.sigmoid(convRCW)) + output * tf.sigmoid(convRCW)
            else:
                output = lay_in + output

    return output

# returns [0,1] tensor: 1 where input not activated (value =< 0), looks at last dimension / features
def zeroes(input :tf.Tensor) -> tf.Tensor:
    axes = [ix for ix in range(len(input.shape))][:-1]      # all but last(feats) axes indexes list like: [0,1,2] for 4d shape
    activated = tf.where(                                   # 1 for value greater than zero, other 0
        condition=  tf.math.greater(input, 0),
        x=          tf.ones_like(input),                    # true
        y=          tf.zeros_like(input))                   # false
    activated_reduced = tf.reduce_sum(activated, axis=axes) # 1 or more for activated, 0 for not activated
    not_activated = tf.equal(activated_reduced, 0)          # true where summed gives zero (~invert)
    nn_zeros = tf.cast(not_activated, dtype=tf.int8)        # cast to int
    return nn_zeros

# dense layer
def lay_dense(
        input,
        units :int,                 # layer width
        name=           'dense',
        reuse=          False,
        activation=     None,
        use_bias=       True,
        initializer=    None,
        seed=           12321):

    if initializer is None: initializer = my_initializer(seed)
    dense_lay = tf.layers.Dense(
        units=              units,
        activation=         activation,
        use_bias=           use_bias,
        kernel_initializer= initializer,
        name=               name,
        _reuse=             reuse)
    output = dense_lay(input)
    return output

# 1d convolution layer, with Gated Linear Unit option
def lay_conv1D(
        input,
        name=           'conv1D',
        kernels=        (3,5,7),        # layer kernels
        filters=        (36,12,6),      # int divisible by len(kernels) or tuple of len(kernels)
        dilation=       1,
        activation=     None,
        use_bias=       True,
        gated_LU=       False,          # Gated Linear Unit architecture
        initializer=    None,
        padding=        'valid',        # 'same' adds padding, 'valid' does not
        seed=           12321,
        verb=           0):

    if initializer is None: initializer = my_initializer(seed)
    with tf.variable_scope(name):
        sub_out_list = []
        if type(kernels) is not tuple: kernels = (kernels,)
        if verb > 1:
            print(' > %s: kernels %s, filters %s, dilation %s' % (name, kernels, filters, dilation))
        for k in range(len(kernels)):
            with tf.variable_scope('kernel_%d' % k):
                sub_kernel = kernels[k]
                if type(filters) is not tuple:  sub_filters = filters // len(kernels)
                else:                           sub_filters = filters[k]
                if gated_LU: sub_filters *= 2

                conv_lay = tf.layers.Conv1D(
                    filters=            sub_filters,
                    kernel_size=        sub_kernel,
                    dilation_rate=      dilation,
                    activation=         None,
                    use_bias=           use_bias,
                    kernel_initializer= initializer,
                    padding=            padding,
                    data_format=        'channels_last')
                sub_output = conv_lay(input)

                if verb > 1: print(' >> sub_conv: filters %s, kernel %s' % (sub_filters, sub_kernel))
                sub_out_list.append(sub_output)

        output = tf.concat(sub_out_list, axis=-1)
        if gated_LU:
            s1, s2 = tf.split(output, num_or_size_splits=2, axis=-1)
            output = s1 * tf.sigmoid(s2)
        elif activation: output = activation(output)

    return output

# 2d convolution layer
def lay_conv2D(
        input,
        name=           'conv2d',
        kernels=        (3, 5, 7),      # layer kernels
        filters=        (36, 12, 6),    # int divisible by len(kernels) or tuple of len(kernels)
        dilation=       1,
        activation=     None,
        useBias=        True,
        gatedLU=        False,          # Gated Linear Unit architecture
        initializer=    None,
        seed=           12321,
        verbLev=        0):

    if initializer is None: initializer = my_initializer(seed)
    with tf.variable_scope(name):
        variables = []
        subOutList = []
        if type(kernels) is not tuple: kernels = (kernels,)
        if verbLev > 0:
            print(' > %s: kernels %s, filetrs %s, dilation %s' % (name, kernels, filters, dilation))
        for k in range(len(kernels)):
            with tf.variable_scope('kernel_%d' % k):
                subKernel = kernels[k]
                if type(filters) is not tuple:
                    subFilters = filters / len(kernels)
                else:
                    subFilters = filters[k]
                if gatedLU: subFilters *= 2

                convLay = tf.layers.Conv2D(
                    filters=            subFilters,
                    kernel_size=        subKernel,
                    dilation_rate=      dilation,
                    activation=         None,
                    use_bias=           useBias,
                    kernel_initializer= initializer,
                    padding=            'valid',
                    data_format=        'channels_last')
                subOutput = convLay(input)
                for var in convLay.variables: variables.append(var)

                if verbLev > 1: print(' >> subConv: filters %s, kernel %s' % (subFilters, subKernel))
                subOutList.append(subOutput)

        output = tf.concat(subOutList, axis=-1)
        if gatedLU:
            s1, s2 = tf.split(output, num_or_size_splits=2, axis=-1)
            output = s1 * tf.sigmoid(s2)
        else:
            if activation: output = activation(output)

        variables = flatten_LOTens(variables)

    return output, variables

# attention for Query, Key and Value
def attn(
        q,                  # decides about output shape
        k,
        v,
        dropout=    0.0,
        drop_flag=  None,
        seed=       12321):

    w = tf.matmul(q, k, transpose_b=True)                   # q*kT, here we calculate weights - how much each key is relevant to each query
    w = w * tf.rsqrt(tf.cast(v.shape[-1].value, w.dtype))   # scale by 1/sqrt(v.dim[-1])
    w = tf.nn.softmax(w)                                    # normalize sum to 1
    if dropout:
        w = tf.layers.dropout(
            inputs=     w,
            rate=       dropout,
            training=   drop_flag,
            seed=       seed)
    att = tf.matmul(w, v)

    return {
        'attention':    att,
        'att_weights':  w}

# time & feats dropout (for sequences)
def tf_drop(
        input,              # tensor [batch,seq,feats]
        time_drop :float,
        feat_drop :float,
        train_flag,         # tensor
        seed=       12321):

    output = input
    in_shape = tf.shape(input)

    # time (per vector) dropout
    if time_drop:
        t_drop = tf.ones(shape=in_shape[-2])
        t_drop = tf.layers.dropout(
            inputs=     t_drop,
            rate=       time_drop,
            training=   train_flag,
            seed=       seed)
        t_drop = tf.expand_dims(t_drop, axis=-1)
        output *= t_drop

    # feature (constant in time) dropout
    if feat_drop:
        f_drop = tf.ones(shape=in_shape[-1])
        f_drop = tf.layers.dropout(
            inputs=     f_drop,
            rate=       feat_drop,
            training=   train_flag,
            seed=       seed)
        f_drop = tf.expand_dims(f_drop, axis=-2)
        output *= f_drop

    return output

# positional encoding layer
def positional_encoding(
        positions :int,         # max number of positions to encode
        width :int,             # width of positions vector
        min_pi_range=       1,
        max_pi_range=       10,
        as_numpy=           True,
        verb=               0):

    angle_rates = np.linspace(min_pi_range/max_pi_range, 1, num=width)
    if verb > 0: print(f'\ni.linspace\n{angle_rates}')
    angle_rates = angle_rates[np.newaxis, :]
    if verb > 0: print(f'\nangle_rates.new_axis\n{angle_rates}')

    pos = np.arange(positions)[:, np.newaxis]
    if verb > 0: print(f'\npos.arange.newaxis\n{pos}')
    pos = pos / positions * max_pi_range
    if verb > 0: print(f'\npos.scaled to range\n{pos}')
    angle_rads = pos * angle_rates
    if verb > 0: print(f'\nangle_rads {angle_rads.shape}\n{angle_rads}')

    angle_rads[:, 0::2] = np.sin(angle_rads[:, 0::2] * np.pi)
    angle_rads[:, 1::2] = np.cos(angle_rads[:, 1::2] * np.pi)

    pos_encoding = angle_rads[np.newaxis, ...]

    pos_encoding = pos_encoding - pos_encoding.mean()

    if as_numpy: pos_encoding = pos_encoding.astype(dtype=np.float32)
    else: pos_encoding = tf.cast(pos_encoding, dtype=tf.float32)
    return pos_encoding


def attention_example():
    print('self attention')
    q = tf.constant(value=np.random.rand(5,4))
    k = tf.constant(value=np.random.rand(5,4))
    v = tf.constant(value=np.random.rand(5,4))
    attn_out = attn(q,k,v)
    print(attn_out)

    print('task attention')
    q = tf.constant(value=np.random.rand(1, 4))
    k = tf.constant(value=np.random.rand(5, 4))
    v = tf.constant(value=np.random.rand(5, 4))
    attn_out = attn(q, k, v)
    print(attn_out)

    print('general attention')
    vector_width = 4
    number_of_queries = 2           # number of queries may vary
    number_of_keys_and_values = 5   # each key is for each value, so their number has to match
    q = tf.constant(value=np.random.rand(number_of_queries,         vector_width))
    k = tf.constant(value=np.random.rand(number_of_keys_and_values, vector_width))
    v = tf.constant(value=np.random.rand(number_of_keys_and_values, vector_width))
    attn_out = attn(q, k, v)
    print(attn_out)


def tf_drop_example():
    v = tf.constant(value=np.random.rand(2,3,4).astype(np.float32))
    print(v)
    v_drop = tf_drop(input=v, time_drop=0.0, feat_drop=0.5, train_flag=True, seed=112)
    print(v_drop)
    with tf.Session() as sess:
        v, v_drop = sess.run([v, v_drop])
        print(v)
        print(v_drop)


if __name__ == '__main__':
    #attention_example()
    tf_drop_example()



"""

 2022 (c) piteren

    RL Trainer for Actor & Critic acting on RLEnvy

        Implements Actor & Critic update

"""

import numpy as np

from pypaq.R4C.helpers import extract_from_batch, RLException
from pypaq.R4C.policy_gradients.base.tf_based.pg_TF_actor import PG_TFActor
from pypaq.R4C.policy_gradients.actor_critic.tf_based.ac_TF_critic import AC_TFCritic
from pypaq.R4C.policy_gradients.pg_trainer import PGTrainer


class ACTrainer(PGTrainer):

    def __init__(
            self,
            actor: PG_TFActor,
            critic_class: type(AC_TFCritic),
            **kwargs):

        # split kwargs assuming that Critic kwargs starts with 'critic_'
        c_kwargs = {k[7:]: kwargs[k] for k in kwargs if k.startswith('critic_')}
        for k in c_kwargs: kwargs.pop(f'critic_{k}')

        PGTrainer.__init__(
            self,
            actor=  actor,
            **kwargs)
        self.actor = actor # INFO: type "upgrade" for pycharm editor

        self.critic = critic_class(
            envy=   kwargs['envy'],
            seed=   kwargs['seed'],
            **c_kwargs)

        self._rlog.info('*** ACTrainer *** initialized')
        self._rlog.info(f'> critic: {critic_class.__name__}')

    # converts one dim arr of ints into two dim one-hot array
    def _actions_OH_encoding(self, actions:np.array) -> np.ndarray:
        hot = np.zeros((len(actions), self.envy.num_actions()))
        hot[np.arange(len(actions)), actions] = 1
        return hot

    # sets terminal states QVs to zeroes
    def _update_terminal_QVs(self, qvs, terminals):
        zeroes = np.zeros(self.envy.num_actions())
        for i in range(len(terminals)):
            if terminals[i]:
                qvs[i] = zeroes
        return qvs

    # update is performed for both: Actor & Critic
    def _update_actor(self, inspect=False) -> dict:

        # TODO: replace prints with logger

        batch = self.memory.get_all()
        self.memory.clear()

        observations =          np.array(extract_from_batch(batch, 'observation'))
        actions =               np.array(extract_from_batch(batch, 'action'))
        rewards =               np.array(extract_from_batch(batch, 'reward'))
        next_observations =              extract_from_batch(batch, 'next_observation')
        terminals =             np.array(extract_from_batch(batch, 'terminal'))

        if inspect:
            print(f'\nBatch size: {len(batch)}')
            print(f'observations: {observations.shape}, {observations[0]}')
            print(f'actions: {actions.shape}, {actions[0]}')
            print(f'rewards {rewards.shape}, {rewards[0]}')
            #print(f'next_observations {next_observations.shape}, {next_observations[0]}')
            print(f'terminals {terminals.shape}, {terminals[0]}')

        # get next_observations actions_probs (with Actor policy)
        next_actions_probs = self.actor.get_policy_probs_batch(next_observations)
        if inspect: print(f'next_actions_probs {next_actions_probs.shape}, {next_actions_probs[0]}')

        # get QVs of current observations
        qvss = self.critic.get_qvs_batch(observations)
        qv_actions = qvss[np.arange(actions.shape[-1]),actions] # get QV of selected actions
        if inspect:
            print(f'qvss {qvss.shape}, {qvss[0]}')
            print(f'qv_actions {qv_actions.shape}, {qv_actions[0]}')

        # update Actor
        act_metrics = self.actor.update_with_experience(
            observations=   observations,
            actions=        actions,
            dreturns=       qv_actions)

        actions_OH = self._actions_OH_encoding(actions)
        if inspect: print(f'actions_OH {actions_OH.shape}, {actions_OH[0]}')

        # get QVs of next observations
        next_actions_qvs = self.critic.get_qvs_batch(next_observations)
        next_actions_qvs = self._update_terminal_QVs(next_actions_qvs, terminals=terminals)
        if inspect: print(f'next_action_qvs {next_actions_qvs.shape}, {next_actions_qvs[0]}')

        # update Critic
        crt_metrics = self.critic.update_with_experience(
            observations=       observations,
            actions_OH=         actions_OH,
            next_action_qvs=    next_actions_qvs,
            next_actions_probs= next_actions_probs,
            rewards=            rewards)
        for k in crt_metrics:
            act_metrics[f'critic_{k}'] = crt_metrics[k]

        if np.isnan(act_metrics['loss']) or np.isnan(crt_metrics['loss']): raise RLException('NaN loss!')

        return act_metrics


from abc import ABC
import numpy as np
from typing import Optional, Callable, List

from pypaq.R4C.policy_gradients.pg_actor import PGActor
from pypaq.R4C.policy_gradients.base.tf_based.pg_TF_graph import pga_graph
from pypaq.neuralmess.nemodel import NEModel



class PG_TFActor(PGActor, ABC):

    def __init__(
            self,
            name: str=                      'PG_TFActor',
            nngraph: Optional[Callable]=    pga_graph,
            **kwargs):
        PGActor.__init__(
            self,
            name=       name,
            nnwrap=     NEModel,
            nngraph=    nngraph,
            **kwargs)

    def get_policy_probs(self, observation: object) -> np.ndarray:
        obs_vec = self._get_observation_vec(observation)
        out = self.nnw(
            feed_dict=  {self.nnw['observation_PH']: [obs_vec]},
            fetch=      ['probs'])
        return out['probs'][0] # reduce dim

    # optimized with batch call to NN
    def get_policy_probs_batch(self, observations: List[object]) -> np.ndarray:
        obs_vecs = self._get_observation_vec_batch(observations)
        out = self.nnw(
            feed_dict=  {self.nnw['observation_PH']: obs_vecs},
            fetch=      ['probs'])
        return out['probs']

    # updates self NN with batch of data
    def update_with_experience(
            self,
            observations,
            actions,
            dreturns,
            inspect=    False) -> dict:
        obs_vecs = self._get_observation_vec_batch(observations)
        out = self.nnw.backward(
            feed_dict=  {
                self.nnw['observation_PH']:  obs_vecs,
                self.nnw['action_PH']:       actions,
                self.nnw['return_PH']:       dreturns},
            fetch=      ['optimizer','probs','loss','gg_norm','gg_avt_norm','actor_ce_mean','zeroes'])
        out.pop('optimizer')
        return out


"""

 2018 (c) piteren

    PaSpa - Parameters Space

        PaSpa is build from PSDD, it has:
            - dim  - dimensionality (= number of axes)
            - rdim - reduced dimensionality (rdim<=dim - since some axes are simpler(tuples, lists of ints))

        each axis of space has:
            - type (list - continuous, tuple - discrete)
            - range (width)

        PaSpa is a metric space (https://en.wikipedia.org/wiki/Metric_space)
        PaSpa supports L1 and L2 distance calculations, normalized to 1 (1 is the longest diagonal of space)

"""

import math
import random
from typing import Optional

from pypaq.pms.base_types import P_VAL, POINT, PSDD

NO_REF = '__NO-REF__'


# Parameters Space
class PaSpa:

    def __init__(
            self,
            psdd : PSDD,            # params space dictionary
            distance_L2=    True,   # sets L1 or L2 distance for PaSpa
            verb=           0):

        self.verb = verb
        self.__psdd = psdd
        self.axes = sorted(list(self.__psdd.keys()))
        self.L2 = distance_L2
        self.dim = len(self.__psdd)
        if self.verb>0:  print(f'\n*** PaSpa (dim: {self.dim}) inits..')

        # resolve axis type and width and some safety checks
        self.__axT = {} # axis type, [list,tuple]_[float,int,diff] list_diff is not allowed
        self.__axW = {} # axis width
        for axis in self.axes:

            if not type(self.__psdd[axis]) in (list,tuple): self.__psdd[axis] = (self.__psdd[axis],) # replace P_VAL by (P_VAL,)

            pdef = self.__psdd[axis]
            tp = 'tuple' if type(pdef) is tuple else 'list'
            assert not (tp == 'list' and len(set(pdef))!= 2), f'ERR: parameter definition with list should have two different elements: >{pdef}<!'

            tpn = 'int' # default
            are_flt = False
            are_dif = False
            for el in pdef:
                if type(el) is float: are_flt = True
                if type(el) is not int and type(el) is not float: are_dif = True
            if are_flt: tpn = 'float'  # downgrade
            if are_dif: tpn = 'diff'   # downgrade

            assert not (tp=='list' and tpn=='diff'), f'ERR: axis {axis} defined with list may contain only floats or ints'

            # sort numeric
            if tpn != 'diff':
                pdef = sorted(list(pdef))
                self.__psdd[axis] = pdef if tp == 'list' else tuple(pdef) # rollback type of sorted

            self.__axT[axis] = f'{tp}_{tpn}' # string like 'list_int'
            self.__axW[axis] = pdef[-1] - pdef[0] if tpn != 'diff' else len(pdef) - 1 # range, for diff_tuple - number of elements

        self.rdim = self.get_rdim()
        if self.verb>0:  print(f' > rdim: {self.rdim:.1f}')

        # width of str(value) for axes, used for str formatting
        self.__str_width = {}
        for axis in self.axes:
            width = max([len(str(e)) for e in self.__psdd[axis]])
            if 'list_float' in self.__axT[axis]: width = 7
            self.__str_width[axis] = width

    # returns copy of self.__psdd
    def get_psdd(self) -> PSDD:
        psdd = {}
        psdd.update(self.__psdd)
        return psdd

    # calculates reduced dimensionality of PaSpa
    def get_rdim(self):
        """
        rdim = log10(∏ sq if sq<10 else 10) for all axes
            sq = 10 for list of floats (axis)
            sq = sqrt(len(axis_elements)) for tuple or list of int
        """
        axd = []
        for axis in self.axes:
            axt = self.__axT[axis]
            if 'list' in axt: sq = 10 if 'float' in axt else math.sqrt(self.__axW[axis])
            else:             sq = math.sqrt(len(self.__psdd[axis]))
            axd.append(sq if sq<10 else 10)
        mul = 1
        for e in axd: mul *= e
        return math.log10(mul)

    # select value from axis closest to given val
    def __closest(
            self,
            ref_val: float or int,      # for diff_tuple it is index of value in self.__psdd[axis]
            axis: str) -> float or int:

        val = ref_val  # for list_float type

        axT = self.__axT[axis]

        if axT == 'list_int': val = int(round(val))

        if 'tuple' in axT:

            # num tuple
            if 'diff' not in axT:
                axis_dst = [(e, abs(e - ref_val)) for e in self.__psdd[axis]]
                axis_dst.sort(key=lambda x: x[1])
                val = axis_dst[0][0]
            # diff_tuple - value of closest index
            else: val = self.__psdd[axis][int(round(val))]

        assert self.__value_in_axis(val, axis)
        return val

    # applies noise to given ref_val (samples new value from sub-range defined by noise scale)
    @staticmethod
    def __apply_noise(
            ref_val: float or int,  # reference value (in range og L & R)
            noise_scale: float,     # <0.0;1.0> distance from ref_val as a factor of range where new val will be sampled
            rngL: float or int,     # range Left value
            rngR: float or int      # range Right value
    ) -> float:
        assert rngL <= ref_val <= rngR, f'ERR: ref_val: {ref_val} not in range: [{rngL};{rngR}]' # TODO: safety check, remove later
        dist = noise_scale * (rngR-rngL)
        # left side of ref_val
        if random.random() < 0.5:
            rng = ref_val - dist        # new left range
            if rng < rngL: rng = rngL   # limit to given range
        # right size
        else:
            rng = ref_val + dist        # new right range
            if rng > rngR: rng = rngR   # limit to given range
        return random.uniform(ref_val, rng)

    # samples random value from whole axis
    def __random_value_noref(self, axis: str) -> P_VAL:

        axT = self.__axT[axis]
        psdd = self.__psdd[axis]

        if 'list' in axT:
            a = psdd[0]
            b = psdd[1]
            val = random.uniform(a,b)
            if 'int' in axT: val = int(round(val))
        else: val = random.choice(psdd)

        assert self.__value_in_axis(value=val, axis=axis) # TODO: safety check, remove later
        return val

    # gets random value for axis (algorithm ensures equal probability for both sides of ref_val)
    def __random_value(
            self,
            axis: str,
            ref_val: P_VAL or NO_REF,   # reference value on axis
            prob_noise: float,          # probability of shifting value with noise (for list or tuple_with_num)
            noise_scale: float,         # max noise scale (axis width, for list or tuple_with_num)
            prob_axis: float,           # probability of value replacement by one sampled from whole axis (for list or num_tuple)
            prob_diff_axis:float        # probability of value replacement by one sampled from whole axis (for diff_tuple)
    ) -> P_VAL:

        if ref_val == NO_REF: return self.__random_value_noref(axis)

        axT = self.__axT[axis]
        axis_psd = self.__psdd[axis]

        # list or num_tuple
        if 'diff' not in axT:

            # whole axis
            if random.random() < prob_axis:
                return self.__random_value_noref(axis)

            val = ref_val
            # add noise
            if noise_scale and random.random() < prob_noise:
                val = PaSpa.__apply_noise(
                    ref_val=        ref_val,
                    noise_scale=    noise_scale,
                    rngL=           axis_psd[0],
                    rngR=           axis_psd[-1])
                val = self.__closest(val, axis)

        # tuple_diff
        else:

            # whole axis
            if random.random() < prob_diff_axis:
                return self.__random_value_noref(axis)

            val = ref_val
            # add noise (on indexes)
            if noise_scale and random.random() < prob_noise:
                val_ix = axis_psd.index(val)
                val_ix = PaSpa.__apply_noise(
                        ref_val=        val_ix,
                        noise_scale=    noise_scale,
                        rngL=           0,
                        rngR=           len(axis_psd)-1)
                val = self.__closest(val_ix, axis)

        assert self.__value_in_axis(value=val, axis=axis) # TODO: safety check, remove later
        return val

    # samples (random) point from whole space or from the surroundings of ref_point
    def __sample_point(
            self,
            ref_point: POINT or NO_REF,         # reference point
            prob_noise=                 0.3,    # probability of shifting value with noise (for list or num_tuple)
            noise_scale=                0.1,    # max noise scale (axis width)
            prob_axis=                  0.1,    # probability of value replacement by one sampled from whole axis (for list or num_tuple)
            prob_diff_axis=             0.3     # probability of value replacement by one sampled from whole axis (for diff_tuple)
    ) -> POINT:
        if ref_point == NO_REF: ref_point = {axis: NO_REF for axis in self.axes}
        return {axis: self.__random_value(
            axis=           axis,
            ref_val=        ref_point[axis],
            prob_noise=     prob_noise,
            noise_scale=    noise_scale,
            prob_axis=      prob_axis,
            prob_diff_axis= prob_diff_axis) for axis in self.axes}

    # samples GX point from given none, one or two
    def sample_point_GX(
            self,
            point_main: Optional[POINT]=    None,   # main parent, when not given > samples from the whole space
            point_scnd: Optional[POINT]=    None,   # scnd parent, when not given > samples from the surroundings of point_main
            prob_mix=                       0.5,    # probability of mixing two values (for list or num_tuple - but only when both parents are given)
            prob_noise=                     0.3,    # probability of shifting value with noise (for list or num_tuple)
            noise_scale=                    0.2,    # max noise scale (axis width)
            prob_axis=                      0.1,    # probability of value replacement by one sampled from whole axis (for list or num_tuple)
            prob_diff_axis=                 0.3     # probability of value replacement by one sampled from whole axis (for diff_tuple)
    ) -> POINT:

        ref_point = NO_REF # for NO_REF samples from whole space

        if point_main:
            assert self.point_from_space(point_main), f'ERR: point_main: {point_main} not from space: {self.__psdd}'

            ref_point = point_main  # select main parent

            # build ref_point with mix or select
            if point_scnd:
                assert self.point_from_space(point_scnd), f'ERR: point_scnd: {point_scnd} not from space: {self.__psdd}'

                ref_point = {}
                ratio = 0.5 + random.random() / 2  # mix/select ratio of parent_main, rest from parent_scnd (at least half from parent_main)

                for axis in self.axes:

                    # mix
                    if random.random() < prob_mix:

                        # num
                        if 'diff' not in self.__axT[axis]:
                            val = point_main[axis] * ratio + point_scnd[axis] * (1-ratio)
                            val = self.__closest(val, axis)
                        # diff_tuple - mix on indexes
                        else:
                            pm_ix = self.__psdd[axis].index(point_main[axis])
                            ps_ix = self.__psdd[axis].index(point_scnd[axis])
                            val_ix = pm_ix * ratio + ps_ix * (1-ratio)
                            val = self.__closest(val_ix, axis)
                    
                    # select
                    else: val = point_main[axis] if random.random() < ratio else point_scnd[axis]

                    ref_point[axis] = val

        # sample from the surroundings of ref_point
        return self.__sample_point(
            ref_point=      ref_point,
            prob_noise=     prob_noise,
            noise_scale=    noise_scale,
            prob_axis=      prob_axis,
            prob_diff_axis= prob_diff_axis)

    # samples 2 corner points with max distance
    def sample_corners(self) -> (POINT, POINT):
        pa = {}
        pb = {}
        left = [0 if random.random()>0.5 else 1 for _ in range(self.dim)] # left/right
        for aIX in range(self.dim):
            ax = self.axes[aIX]
            vl = self.__psdd[ax][0]
            vr = self.__psdd[ax][-1]
            pa[ax] = vl
            pb[ax] = vr
            if left[aIX]:
                pa[ax] = vr
                pb[ax] = vl
        return pa, pb

    # checks if given value belongs to an axis of this space
    def __value_in_axis(self, value: P_VAL, axis: str) -> bool:
        if axis not in self.__axT:                                      return False # axis not in a space
        if 'list' in self.__axT[axis]:
            if type(value) is float and 'int' in self.__axT[axis]:      return False # type mismatch
            if self.__psdd[axis][0] <= value <= self.__psdd[axis][1]:   return True
            else:                                                       return False # value not in a range
        elif value not in self.__psdd[axis]:                            return False # value not in a tuple
        return True

    # checks if point comes from a space (is built from same axes and is in space and)
    def point_from_space(self, point: POINT) -> bool:
        if set(point.keys()) != set(self.axes): return False
        for axis in point:
            if not self.__value_in_axis(value=point[axis], axis=axis):
                return False
        return True

    # distance between two points in this space (normalized to 1 - divided by max space distance)
    def distance(self, pa: POINT, pb: POINT) -> float:

        if self.L2:
            dist_pow_sum = 0
            for axis in pa:
                if self.__axW[axis] > 0:
                    dist = self.__psdd[axis].index(pa[axis]) - self.__psdd[axis].index(pb[axis]) \
                        if 'diff' in self.__axT[axis] else \
                        pa[axis] - pb[axis]
                    dist_pow_sum += (dist / self.__axW[axis])**2
            return  math.sqrt(dist_pow_sum) / math.sqrt(self.dim)
        else:
            dist_abs_sum = 0
            for axis in pa:
                if self.__axW[axis] > 0:
                    dist = self.__psdd[axis].index(pa[axis]) - self.__psdd[axis].index(pb[axis]) \
                        if 'diff' in self.__axT[axis] else \
                        pa[axis] - pb[axis]
                    dist_abs_sum += abs(dist) / self.__axW[axis]
            return dist_abs_sum / self.dim

    # merges two PSDD
    @staticmethod
    def merge_psdd(
            psdd_a: PSDD,
            psdd_b: PSDD) -> PSDD:
        return (PaSpa(psdd_a) + PaSpa(psdd_b)).get_psdd()

    # returns info(string) about self
    def __str__(self):
        info = f'*** PaSpa *** (dim: {self.dim}, rdim: {self.rdim:.1f}) parameters space:\n'
        max_ax_l = 0
        max_ps_l = 0
        for axis in self.axes:
            if len(axis)                > max_ax_l: max_ax_l = len(axis)
            if len(str(self.__psdd[axis])) > max_ps_l: max_ps_l = len(str(self.__psdd[axis]))
        if max_ax_l > 40: max_ax_l = 40
        if max_ps_l > 40: max_ps_l = 40

        for axis in self.axes:
            info += f' > {axis:{max_ax_l}s}  {str(self.__psdd[axis]):{max_ps_l}s}  {self.__axT[axis]:11s}  width: {self.__axW[axis]}\n'
        return info[:-1]

    # same axes, same definitions, same L
    def __eq__(self, other):
        if self.axes != other.axes: return False
        for k in self.__psdd.keys():
            if self.__psdd[k] != other.__psdd[k]:
                return False
        if self.L2 != other.L2:
            return False
        return True

    def __add__(self, other: "PaSpa") -> "PaSpa":
        psdd_a = self.get_psdd()
        psdd_b = other.get_psdd()
        psdd_merged = {}
        psdd_merged.update(psdd_a)
        for ax in other.axes:
            if ax not in psdd_merged: psdd_merged[ax] = psdd_b[ax]
            else:
                assert type(psdd_merged[ax]) == type(psdd_b[ax]), f'ERR: types of axes {ax} for PSDD in both PaSpa do not match!'
                if 'tuple' in other.__axT[ax]:
                    # add new elements to the end
                    if 'diff' in other.__axT[ax]:
                        merged = list(psdd_merged[ax])
                        for e in psdd_b[ax]:
                            if e not in merged: merged.append(e)
                    else:
                        merged = list(set(list(psdd_merged[ax]) + list(psdd_b[ax])))
                        merged.sort()
                    psdd_merged[ax] = tuple(merged)
                else:
                    ranges = psdd_merged[ax] + psdd_b[ax]
                    psdd_merged[ax] = [min(ranges), max(ranges)]
        return PaSpa(psdd_merged)


from pypaq.neuralmess.base_elements import tf
from pypaq.neuralmess.layers import lay_dense

def critic_graph(
        name=               'ac_critic',
        observation_width=  4,
        gamma=              0.99, # discount factor (gamma)
        num_actions=        2,
        hidden_layers=      (24,24),
        lay_norm=           False,
        seed=               123,
        verb=               0):

    with tf.variable_scope(name):

        observation_PH = tf.placeholder(
            shape=  (None, observation_width),
            dtype=  tf.float32,
            name=   'observation')
        action_OH_PH = tf.placeholder( # one-hot vector of action taken
            shape=  (None, num_actions),
            dtype=  tf.float32,
            name=   'action_OH_probs')
        next_action_qvs_PH = tf.placeholder(
            shape=  (None, num_actions),
            dtype=  tf.float32,
            name=   'next_action_qvs')
        next_action_probs_PH = tf.placeholder(
            shape=  (None, num_actions),
            dtype=  tf.float32,
            name=   'next_action_probs')
        reward_PH = tf.placeholder(
            shape=  None,
            dtype=  tf.float32,
            name=   'reward')

        layer = observation_PH
        if lay_norm: layer = tf.keras.layers.LayerNormalization(axis=-1)(layer)
        for i in range(len(hidden_layers)):
            layer = lay_dense(
                input=      layer,
                name=       f'hidden_layer_{i}',
                units=      hidden_layers[i],
                activation= tf.nn.relu,
                seed=       seed)
            if lay_norm: layer = tf.keras.layers.LayerNormalization(axis=-1)(layer)

        # QVs (for all actions of current observation)
        qvs = lay_dense(
            input=      layer,
            name=       'qvs',
            units=      num_actions,
            activation= None,
            seed=       seed)

        qv =     tf.reduce_sum(qvs *                action_OH_PH,         axis=1) # Q(s,a)
        next_V = tf.reduce_sum(next_action_qvs_PH * next_action_probs_PH, axis=1) # V(next_s)
        labels = reward_PH + gamma * next_V
        loss = tf.reduce_mean(tf.losses.mean_squared_error(predictions=qv, labels=labels))
        if verb>0:
            print(f' > qvs: {qvs}')
            print(f' > qv: {qv}')
            print(f' > labels: {labels}')

    return {
        'observation_PH':       observation_PH,
        'action_OH_PH':         action_OH_PH,
        'next_action_qvs_PH':   next_action_qvs_PH,
        'next_action_probs_PH': next_action_probs_PH,
        'reward_PH':            reward_PH,
        'qvs':                  qvs,
        'loss':                 loss}


import logging
from typing import Optional

from pypaq.lipytools.little_methods import stamp
from pypaq.lipytools.files import prep_folder


# returns formatted Logger
def get_pylogger(
        name: Optional[str]=    None,
        add_stamp=              True,   # to prevent merging loggers from one class many objects
        folder: Optional[str]=  None,   # if given then writes logfile
        level=                  logging.INFO,
        format: str=            '%(asctime)s {%(filename)17s:%(lineno)3d} p%(process)s %(levelname)s: %(message)s',
        to_stdout=              True):

    if not name: name = 'logger'
    if add_stamp: name += '_' + stamp()

    formatter = logging.Formatter(format)

    # manage file handler
    fh = None
    if folder:
        prep_folder(folder)
        fh = logging.FileHandler(f'{folder}/{name}.log')
        fh.setFormatter(formatter)

    # manage stream handler
    sh = None
    if to_stdout:
        sh = logging.StreamHandler()
        sh.setFormatter(formatter)

    logger = logging.getLogger(name)
    logger.setLevel(level)
    if fh: logger.addHandler(fh)
    if sh: logger.addHandler(sh)
    return logger

# returns child with higher level
def get_hi_child(
        logger,
        name: Optional[str]=    None,
        higher_level=           True):
    if not name: name = '_child'
    clogger = logger.getChild(name)
    if higher_level:
        lvl = clogger.getEffectiveLevel()
        clogger.setLevel(lvl+10)
    return clogger


"""

 2022 (c) piteren

    baseline DQN Module for MOTorch (PyTorch)

"""

import torch
from typing import Dict

from pypaq.torchness.motorch import Module
from pypaq.torchness.layers import LayDense


class DQNModel(Module):

    def __init__(
            self,
            num_actions: int=   4,
            observation_width=  4,
            hidden_layers=      (12,),
            seed=               121):

        torch.nn.Module.__init__(self)

        lay_shapeL = []
        next_in = observation_width
        for hl in hidden_layers:
            lay_shapeL.append((next_in,hl))
            next_in = hl

        self.ln = torch.nn.LayerNorm(observation_width) # input layer norm

        self.linL = [LayDense(*shape) for shape in lay_shapeL]
        self.lnL = [torch.nn.LayerNorm(shape[-1]) for shape in lay_shapeL]

        lix = 0
        for lin,ln in zip(self.linL, self.lnL):
            self.add_module(f'lay_lin{lix}', lin)
            self.add_module(f'lay_ln{lix}', ln)
            lix += 1

        self.logits = LayDense(
            in_features=    next_in,
            out_features=   num_actions,
            activation=     None)

        self.loss = torch.nn.MSELoss(reduction='none')

    def forward(self, obs) -> dict:
        out = self.ln(obs)
        for lin,ln in zip(self.linL,self.lnL):
            out = lin(out)
            out = ln(out)
        logits = self.logits(out)
        return {'logits': logits}

    def accuracy(
            self,
            logits: torch.Tensor,
            labels: torch.Tensor) -> float:
        # no good accuracy for this model
        return 0.0

    def loss_acc(self, obs, lbl, mask=None) -> dict:
        out = self(obs)
        logits = out['logits']
        loss = self.loss(logits, lbl)
        if mask is not None: loss *= mask       # mask
        loss = torch.sum(loss, dim=-1)          # reduce over samples
        out['loss'] = torch.mean(loss)          # average
        out['acc'] = self.accuracy(logits, lbl)
        return out


from typing import Optional

# moving average class, updates self.value with factor (given while init)
class MovAvg:

    def __init__(self, factor=0.1):
        self.value: Optional[float] = None
        self.factor = factor

    def upd(self, val: float):
        if self.value is None: self.value = val
        else: self.value = (1-self.factor)*self.value + self.factor*val
        return self.value

    def __call__(self) -> float: return self.value


from collections import OrderedDict
import torch
from typing import Optional


class TorchnessException(Exception):
    pass

# weights initializer from BERT, the only difference is that in torch values are CLAMPED not SAMPLED till in <a,b>
def bert_initializer(*args, std=0.02, **kwargs):
    return torch.nn.init.trunc_normal_(*args, **kwargs, std=std, a=-2*std, b=2*std)

def my_initializer(*args, std=0.02, **kwargs):
    # different layers use different initialization functions:
    # torch Linear % Conv1D uses kaiming_uniform_(weights) & xavier_uniform_(bias)
    # my TF uses trunc_normal_(weights, std=0.02, a==b==2*std) & 0(bias) <- from BERT
    # - kaiming_uniform_ is uniform_ with bound from 2015 paper, (for relu)
    # - xavier_uniform_ is uniform_ whit bound from 2010 paper (for linear / sigmoid)
    # - trunc_normal_ is normal with mean 0 and given std, all values SAMPLED till in <a,b>
    return bert_initializer(*args, **kwargs, std=std)

# weighted merge of two checkpoints, does NOT check for compatibility of two checkpoints, but will crash if those are not compatible
def mrg_ckpts(
        ckptA: str,                     # checkpoint A (file name)
        ckptB: Optional[str],           # checkpoint B (file name), for None takes 100% ckptA
        ckptM: str,                     # checkpoint merged (file name)
        ratio: float=           0.5,    # ratio of merge
        noise: float=           0.0     # noise factor, amount of noise added to new value <0.0;1.0>
):

    checkpoint_A = torch.load(ckptA)
    checkpoint_B = torch.load(ckptB) if ckptB else checkpoint_A

    cmsd_A = checkpoint_A['model_state_dict']
    cmsd_B = checkpoint_B['model_state_dict']
    cmsd_M = OrderedDict()

    # TODO: watch out for not-float-tensors -> such should be taken from A without a mix
    for k in cmsd_A:
        std_dev = float(torch.std(cmsd_A[k]))
        noise_tensor = torch.zeros_like(cmsd_A[k])
        if std_dev != 0.0: # bias variable case
            my_initializer(noise_tensor, std=std_dev)
        cmsd_M[k] = ratio * cmsd_A[k] + (1 - ratio) * cmsd_B[k] + noise * noise_tensor

    checkpoint_M = {}
    checkpoint_M.update(checkpoint_A)
    checkpoint_M['model_state_dict'] = cmsd_M
    torch.save(checkpoint_M, ckptM)


def scaled_cross_entropy(
        labels,
        scale,
        logits: Optional[torch.Tensor]= None,
        probs: Optional[torch.Tensor]=  None) -> torch.Tensor:

    if logits is None and probs is None:
        raise TorchnessException('logits OR probs must be given!')

    if probs is None:
        probs = torch.nn.functional.softmax(input=logits, dim=-1)

    prob_label = probs[range(len(labels)), labels] # probability of class from label

    # merge loss for positive and negative advantage
    ce = torch.where(
        condition=  scale > 0,
        input=      -torch.log(prob_label),
        other=      -torch.log(1-prob_label))

    return ce * torch.abs(scale)


from collections import deque
import random
from typing import Iterable, Sized


class DataSampler:

    def __init__(
            self,
            data: Sized and Iterable,
            seed = 123):
        self.__seed_counter = seed
        self.__data = data
        self.__data_shuffled = self.__get_data_shuffled()

    # prepares new deque of shuffled data
    def __get_data_shuffled(self) -> deque:
        random.seed(self.__seed_counter)
        self.__seed_counter += 1
        indexes = list(range(len(self.__data)))
        random.shuffle(indexes)
        # print(indexes)
        data_shuffled = [self.__data[i] for i in indexes]
        # print(data_shuffled)
        data_shuffled = deque(data_shuffled)
        return data_shuffled

    def get_sample(self):
        if not self.__data_shuffled:
            self.__data_shuffled = self.__get_data_shuffled()
        return self.__data_shuffled.popleft()

    def epoch_finished(self) -> bool:
        return not self.__data_shuffled


if __name__ == '__main__':

    data = [1,2,3,4,5,6,7,8]
    ds = DataSampler(data)
    for _ in range(20):
        print(ds.get_sample())
        if ds.epoch_finished(): print('epoch finished')


def get_tf():

    import os
    os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

    import tensorflow
    tensorflow.compat.v1.logging.set_verbosity(tensorflow.compat.v1.logging.ERROR)
    if str(tensorflow.__version__)[0]=='2':
        import tensorflow.compat.v1 as tf
        tf.disable_v2_behavior()
    else: import tensorflow as tf
    return tf

tf = get_tf()


from torch.utils.tensorboard import SummaryWriter


# TensorBoard writer based on PyTorch wraps of TensorBoard
class TBwr:

    def __init__(
            self,
            logdir: str,
            flush_secs= 10):
        self.logdir = logdir
        self.flush_secs = flush_secs
        # INFO: SummaryWriter creates logdir while init, because of that self.sw init has moved here (in the first call of add)
        self.sw = None

    def _get_sw(self):
        return SummaryWriter(
            log_dir=    self.logdir,
            flush_secs= self.flush_secs)

    def add(self,
            value,
            tag: str,
            step: int):
        if not self.sw: self.sw = self._get_sw()
        self.sw.add_scalar(
            tag=            tag,
            scalar_value=   value,
            global_step=    step)

    def add_histogram(
            self,
            values,
            tag: str,
            step: int):
        if not self.sw: self.sw = self._get_sw()
        self.sw.add_histogram(
            tag=            tag,
            values=         values,
            global_step=    step,
            bins=           "tensorflow")

    def flush(self): self.sw.flush()



"""

 2019 (c) piteren

    NN TF graph elements and tools

"""

import numpy as np
from typing import Optional, List

from pypaq.neuralmess.get_tf import tf
from pypaq.mpython.devices import DevicesParam, mask_cuda_devices
from pypaq.lipytools.little_methods import short_scin
from pypaq.lipytools.pylogger import get_pylogger
from pypaq.mpython.mpdecor import proc_wait


# default initializer for variables of graph
def my_initializer(seed=12321, stddev=0.02):
    #old initializers
    #tf.contrib.layers.xavier_initializer(uniform=False,seed=seed)
    #tf.contrib.layers.variance_scaling_initializer(factor=1.0,mode='FAN_IN',uniform=False,seed=seed) # https://www.tensorflow.org/api_docs/python/tf/contrib/layers/variance_scaling_initializer
    #tf.random_normal_initializer(stddev=0.01, seed=seed)
    #tf.random_uniform_initializer(minval=-0.01, maxval=0.01, seed=seed)

    return tf.truncated_normal_initializer(stddev=stddev, seed=seed)  # recommended by TF, BERT (0.02?)

# GeLU (Gaussian Error Linear Unit) activation https://arxiv.org/abs/1606.08415
def gelu(x):
    cdf = 0.5 * (1.0 + tf.tanh((np.sqrt(2 / np.pi) * (x + 0.044715 * tf.pow(x, 3)))))
    return x * cdf

# scales learning rate with warm-up and annealing (after warm-up)
def lr_scaler(
        baseLR,                     # base learning rate
        g_step: tf.Tensor=  None,   # global step tf.variable of tf.int type, for None creates one
        warm_up: int=       1000,   # warmup steps, None or 0 turns-off
        ann_base: float=    0.999,  # annealing base, None or 1 for turn-off
        ann_step: float=    1.0,    # annealing step, higher value speeds up annealing
        n_wup_off: float=   2.0,    # N warmUp offset of annealing
        logger=             None):

    if not logger: logger = get_pylogger()
    logger.info(f'*** lr_scaler for baseLR: {baseLR}')
    baseLR = tf.convert_to_tensor(baseLR)

    # create global step variable if not given
    if g_step is None:
        g_step = tf.get_variable(
            name=           'g_step',
            shape=          [],
            trainable=      False,
            initializer=    tf.constant_initializer(0),
            dtype=          tf.int32)

    g_step_fl = tf.cast(g_step, dtype=tf.float32)
    if warm_up is None: warm_up = 0
    lR = baseLR
    if warm_up:
        ratioWm = tf.reduce_min([g_step_fl, warm_up]) / warm_up # warmUp ratio
        lR = baseLR * ratioWm # learning rate with warmup
        logger.debug(f'applied warmUp ({warm_up}) to lR')
    if ann_base is not None and ann_base != 1:
        gStep_offs = tf.reduce_max([0, g_step_fl - warm_up * n_wup_off]) # offset by warmUpSteps
        lR *= ann_base ** (gStep_offs * ann_step) # learning rate with annealing
        logger.debug(f'applied annealing to lR ({ann_base:.5f},{ann_step:.5f})')
    return {
        'scaled_LR':    lR,
        'g_step':       g_step}

# gradient clipping function (by clip_value or AVT algorithm)
def grad_clipper_AVT(
        gradients,                  # gradients to clip
        clip_value=     None,       # clipping value, for None clips with avt
        avt_SVal=       0.1,        # start value for AVT (smaller value makes warmup)
        avt_window=     100,        # width of averaging window (number of steps)
        avt_max_upd=    1.5,        # single step max factor of avt update
        do_clip=        True,       # disables clipping (just GN calculations)
        logger=         None):

    if not logger: logger = get_pylogger()

    gg_norm = tf.global_norm(gradients) # gradients global norm
    gg_avt_norm = tf.get_variable( # time averaged gradients global norm variable
        name=           'gg_avt_norm',
        shape=          [],
        trainable=      False,
        initializer=    tf.constant_initializer(avt_SVal),
        dtype=          tf.float32)

    avt_update = tf.reduce_min([gg_norm, avt_max_upd * gg_avt_norm]) # single value to update AVTG with (current GNorm or clipped to max value)
    # assign new value
    gg_avt_norm = tf.assign(
        ref=    gg_avt_norm,
        value=  (gg_avt_norm * (avt_window-1) + avt_update) / avt_window)
    logger.debug(f'grad_clipper_AVT: avt_SVal {avt_SVal:.1f}, avt_window {avt_window}, avt_max_upd {avt_max_upd:.1f}')

    if do_clip:
        gradients, _ = tf.clip_by_global_norm(
            t_list=     gradients,
            clip_norm=  clip_value or gg_avt_norm,
            use_norm=   gg_norm)
        logger.debug(f' >> is clipping gradients {"with value" if clip_value else "with AVT"}')
    else: logger.debug(' >> not doing clipping')

    return {
        'gradients':    gradients,
        'gg_norm':      gg_norm,
        'gg_avt_norm':  gg_avt_norm}

# gradient clipping loss reductor (gradient clipping + optimizer) (wraps grad_clipper_AVT)
def gc_loss_reductor(
        optimizer :tf.compat.v1.train.Optimizer,
        vars :list=         None,
        g_step=             None,   # put here globalStep variable to update +1 with optimizer
        avg_loss=           None,   # put here loss if you do not have gradients yet
        gradients=          None,
        clip_value=         None,
        avt_SVal: float=    0.1,
        avt_window: int=    100,
        avt_max_upd: float= 1.5,
        do_clip: bool=      True,
        logger=             None):

    if not logger: logger = get_pylogger()
    if vars is None: vars = tf.trainable_variables()

    if gradients is None: gradients = tf.gradients(avg_loss, vars, colocate_gradients_with_ops=False)

    gc_out= grad_clipper_AVT(
        gradients=      gradients,
        clip_value=     clip_value,
        avt_SVal=       avt_SVal,
        avt_window=     avt_window,
        avt_max_upd=    avt_max_upd,
        do_clip=        do_clip,
        logger=         logger)
    clippedGradients =  gc_out['gradients']

    optimizer = optimizer.apply_gradients(
        grads_and_vars= zip(clippedGradients, vars),
        global_step=    g_step)

    return {
        'optimizer':    optimizer,
        'gg_norm':      gc_out['gg_norm'],
        'gg_avt_norm':  gc_out['gg_avt_norm']}

# replaces nan values @tensor with zero
def replace_nan_with_zero(tensor): return tf.where(tf.is_nan(tensor), tf.zeros_like(tensor), tensor)

# flattens list of tensors of any dimensions and returns one (flattened) tensor
def flatten_LOTens(tList):

    resh_vars = [tf.reshape(var, [-1]) for var in tList]
    return tf.concat(resh_vars, axis=-1)

# returns list of indexes of n layers, len(list)==n_select
def list_of_layers(
        n_layers :int,          # num of all layers
        n_select :int=  None):  # num of layers to select

    if n_select is None: n_select = n_layers
    if n_select > n_layers: n_select = n_layers

    layers = [] # 0 case
    if n_select == 1: layers = [0]
    if n_select > 1:

        layers = [0, n_layers - 1]  # first and last

        for i in range(n_select - 2):
            layers.append(int((n_layers - 1) / (n_select - 1) * (i + 1)))

    return sorted(list(set(layers)))

# calculates and returns num of variables(floats) @graph from given variables list or scope
def num_var_floats(
        variables :list=    None,   # list of variables
        graph=              None,
        scope :str=         None):  # scope name

    assert variables is not None or (scope and graph), 'ERR: variables list or (scope & graph) must be given'
    if variables is None: variables = graph.get_collection(name='variables', scope=scope)

    #with graph.as_default():
    numVars = sum(v.get_shape().num_elements() for v in variables)
    return numVars

# calculates size (num of values) of tensor given its shape
def sh_size(shape : list or tuple):
    size = 1
    for e in shape: size *= e
    return size

# prints variables log
def log_vars(
        variables: list,
        simple=     False,
        sort=       True) -> str: # use order from list or sorted by name
    s = f'Total num of variables: {len(variables)}\n'
    s += f' > num of floats: {short_scin(num_var_floats(variables), precision=2)}\n'
    if not simple:
        vns = [(v.name, v.shape) for v in variables]
        if sort:
            dVar = {v.name: v.shape for v in variables}
            vns = [(key, dVar[key]) for key in sorted(list(dVar.keys()))]
        for v in vns: s += f' >> v: {v[0]} {v[1]}\n'
    return s[:-1]

# logs variables from checkpoint
def log_checkpoint(ckpt_FD):
    ckpt_vars = tf.train.list_variables(ckpt_FD)
    tot_siz = 0
    for _, shape in ckpt_vars:
        tot_siz += sh_size(shape)
    print(f'\nGot {len(ckpt_vars)} variables in original checkpoint (total size {tot_siz} nums)')

    max_nm_len = 0
    max_sh_len = 0
    for var_name, shape in ckpt_vars:
        if len(var_name) > max_nm_len:   max_nm_len = len(var_name)
        if len(str(shape)) > max_sh_len: max_sh_len = len(str(shape))
    if max_nm_len > 90: max_nm_len = 90
    if max_sh_len > 20: max_sh_len = 20

    for var_name, shape in ckpt_vars:
        var = tf.train.load_variable(ckpt_FD, var_name)
        print(f' > ({100*sh_size(shape)/tot_siz:4.1f}%) {var_name:{max_nm_len}s} {str(shape):{max_sh_len}s} {var.dtype}')

# weighted merge of two checkpoints, does NOT check for compatibility of two checkpoints, but will crash if those are not compatible
@proc_wait
def mrg_ckpts(
        ckptA: str,                     # checkpoint A (folder name)
        ckptA_FD: str,                  # root folder of cpktA (absolute or relative)
        ckptB: str or None,             # checkpoint B (folder name), for None takes 100% ckptA
        ckptB_FD: str or None,          # root folder of cpktB (absolute or relative)
        ckptM: str,                     # checkpoint merged (folder name)
        ckptM_FD: str,                  # root folder of cpktM (absolute or relative)
        ratio: float=           0.5,    # ratio of merge
        noise: float=           0.0,    # noise factor, amount of noise added to new value (0.0-1.0...)
        replace_scope: str=     None,   # replaces outer scope with given string
        devices :DevicesParam=  None,   # by default does merge on CPU
        verb=                   0):

    mask_cuda_devices(devices)

    if ckptA_FD[-1] != '/': ckptA_FD += '/'
    if ckptB_FD and ckptB_FD[-1] != '/': ckptB_FD += '/'
    if ckptM_FD[-1] != '/': ckptM_FD += '/'

    var_namesA = sorted([v[0] for v in tf.train.list_variables(ckptA_FD+ckptA)])
    if verb>0: print(f'variables from ckptA ({len(var_namesA):4d}): {var_namesA}')
    var_namesB = sorted([v[0] for v in tf.train.list_variables(ckptB_FD+ckptB)]) if ckptB else []
    if verb>0: print(f'variables from ckptB ({len(var_namesB):4d}): {var_namesB}')

    oscope_len = 0
    if replace_scope:
        for c in var_namesA[0]:
            if c == '/': break
            oscope_len += 1
    if verb>0:
        print(f'oscope_len {oscope_len}')
        if oscope_len: print(f' > will replace {var_namesA[0][:oscope_len]} with {replace_scope}')

    avL = []
    with tf.variable_scope('av'):
        for var_name in var_namesA:
            var = tf.train.load_variable(f'{ckptA_FD}{ckptA}', var_name)
            avL.append(tf.Variable(var, name=var_name))

    bvL = []
    if ckptB:
        with tf.variable_scope('bv'):
            for var_name in var_namesB:
                var = tf.train.load_variable(f'{ckptB_FD}{ckptB}', var_name)
                bvL.append(tf.Variable(var, name=var_name))

    cvL = []
    for ix in range(len(var_namesA)):
        var_name = var_namesA[ix]
        if verb>0: print(f'old var_name: {var_name}')
        if replace_scope: var_name = replace_scope + var_name[oscope_len:]

        varA = avL[ix]
        if bvL and varA.dtype == 'float32':
            varB = bvL[ix]
            noise_tensor = tf.random.truncated_normal( # random values from normal distribution truncated by 2stddev
                shape=  varA.shape,
                stddev= tf.math.reduce_std(varA)) # stddev of varA
            var = tf.Variable(ratio * varA + (1 - ratio) * varB + noise * noise_tensor, name=var_name)
        else:
            var = tf.Variable(varA, name=var_name)
        cvL.append(var)

    # save
    if verb>0: print('\nWriting checkpoint... ', end='')
    child_saver = tf.train.Saver(cvL)
    #config = tf.ConfigProto()
    #config.gpu_options.allow_growth = True
    with tf.Session(
            #config=config
            ) as sess:
        sess.run(tf.global_variables_initializer())
        child_saver.save(sess, f'{ckptM_FD}{ckptM}/{ckptM}', write_meta_graph=False)
    tf.reset_default_graph()
    if verb>0: print('done!')

# TensorBoard writer
class TBwr:

    def __init__(
            self,
            logdir: str,
            flush_secs= 10):
        self.logdir = logdir
        self.flush_secs = flush_secs
        # INFO: tf.summary.FileWriter creates logdir while init, because of that self.sw init has moved here (in the first call of add)
        self.sw = None

    def add(self,
            value,
            tag: str,
            step: int):

        if not self.sw:
            self.sw = tf.summary.FileWriter(logdir=self.logdir, flush_secs=self.flush_secs)

        sv = tf.Summary(value=[tf.Summary.Value(tag=tag, simple_value=value)])
        self.sw.add_summary(sv, step)

    def add_summary(self, summ, step):
        self.sw.add_summary(summ, step)

    def flush(self): self.sw.flush()


"""

 2022 (c) piteren

    TB (TensorBoard) writer wrapped with separate subprocess

"""

from pypaq.mpython.mptools import ExSubprocess, Que, QMessage

"""
# TensorBoard writer
class TBwr:

    # Internal Process (for TB) with all TF inside (separated from external process)
    class TB_IP(ExSubprocess):

        FLUSH_MSG =  QMessage(type='flush',  data=None)
        POISON_MSG = QMessage(type='poison', data=None)

        def __init__(
                self,
                logdir: str,
                set_to_CPU: bool,
                flush_secs: int):
            ExSubprocess.__init__(self, ique=Que(), oque=Que())
            self.logdir = logdir
            self.set_to_CPU = set_to_CPU
            self.flush_secs = flush_secs

        # method called while starting process (ExSubprocess), processes messages for TB
        def subprocess_method(self):

            import os
            import warnings

            os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
            warnings.filterwarnings("ignore", category=DeprecationWarning)

            import tensorflow as tf

            if self.set_to_CPU:
                # check for TF visible_physical_devices and set only CPU as visible
                visible_physical_devices = tf.config.list_physical_devices()
                set_visible = []
                for dev in visible_physical_devices:
                    if 'CPU' in dev.name:
                        set_visible.append(dev)
                tf.config.set_visible_devices(set_visible)

            # INFO: tf.summary.create_file_writer creates logdir while init, because of that self.sw init has moved to the first call of add()
            sw = None
            while True:

                msg = self.ique.get()

                if msg.type == 'poison': break

                if msg.type == 'flush' and sw: sw.flush()

                if msg.type == 'add':
                    if not sw:
                        sw = tf.summary.create_file_writer(
                            logdir=         self.logdir,
                            flush_millis=   1000 * self.flush_secs)
                    with sw.as_default():
                        tf.summary.scalar(
                            name=   msg.data['tag'],
                            data=   msg.data['value'],
                            step=   msg.data['step'])

        def exit(self):
            self.ique.put(TBwr.TB_IP.POISON_MSG)


    def __init__(
            self,
            logdir: str,
            set_to_CPU=     True,
            flush_secs=     10):
        self.tb_ip = TBwr.TB_IP(
            logdir=     logdir,
            set_to_CPU= set_to_CPU,
            flush_secs= flush_secs)
        self.tb_ip.start()

    def add(self,
            value,
            tag: str,
            step: int):
        data = {'tag':tag, 'value':value, 'step':step}
        self.tb_ip.ique.put(QMessage(type='add', data=data))

    def flush(self):
        self.tb_ip.ique.put(TBwr.TB_IP.FLUSH_MSG)

    def exit(self):
        self.tb_ip.ique.put(TBwr.TB_IP.POISON_MSG)
"""

import tensorflow as tf


# TensorBoard writer
class TBwr:

    def __init__(
            self,
            logdir: str,
            set_to_CPU=     True,
            flush_secs=     10):

        # TODO: is this version CPU-GPU-config compatible??
        # TODO: INFO: tf.summary.create_file_writer creates logdir while init, because of that self.sw init has moved to the first call of add()
        self.sw = tf.summary.create_file_writer(
            logdir=         logdir,
            flush_millis=   1000 * flush_secs)

    def add(self,
            value,
            tag: str,
            step: int):
        with self.sw.as_default():
            tf.summary.scalar(
                name=tag,
                data=value,
                step=step)

    def flush(self):
        self.sw.flush()

    def exit(self):
        pass


import tensorflow as tf

from pypaq.neuralmess_duo.base_elements import my_initializer


# dense layer wrap, wraps with call, adds default initializer
def lay_dense(
        input,
        units :int,                 # layer width
        name=           'dense',
        activation=     None,
        use_bias=       True,
        initializer=    None,
        seed=           12321):
    if initializer is None: initializer = my_initializer(seed)
    dense_lay = tf.keras.layers.Dense(
        units=              units,
        activation=         activation,
        use_bias=           use_bias,
        kernel_initializer= initializer,
        name=               name)
    output = dense_lay(input)
    return output